package workspace

import (
	"errors"

	"cloud.o-forge.io/core/oc-lib/dbs"
	"cloud.o-forge.io/core/oc-lib/dbs/mongo"
	"cloud.o-forge.io/core/oc-lib/models/resources/data"
	"cloud.o-forge.io/core/oc-lib/models/resources/datacenter"
	"cloud.o-forge.io/core/oc-lib/models/resources/processing"
	"cloud.o-forge.io/core/oc-lib/models/resources/storage"
	w "cloud.o-forge.io/core/oc-lib/models/resources/workflow"
	"cloud.o-forge.io/core/oc-lib/models/utils"
)

type workspaceMongoAccessor struct {
	utils.AbstractAccessor
}

func New() *workspaceMongoAccessor {
	return &workspaceMongoAccessor{}
}

func (wfa *workspaceMongoAccessor) DeleteOne(id string) (utils.DBObject, int, error) {
	return wfa.GenericDeleteOne(id, wfa)
}

func (wfa *workspaceMongoAccessor) UpdateOne(set utils.DBObject, id string) (utils.DBObject, int, error) {
	d := set.(*Workspace)
	d.DataResources = nil
	d.DatacenterResources = nil
	d.StorageResources = nil
	d.ProcessingResources = nil
	d.WorkflowResources = nil
	if d.Active {
		res, _, err := wfa.LoadAll()
		if err == nil {
			for _, r := range res {
				if r.GetID() != id {
					r.(*Workspace).Active = false
					wfa.UpdateOne(r.(*Workspace), r.GetID())
				}
			}
		}
	}
	return wfa.GenericUpdateOne(set, id, wfa, &Workspace{})
}

func (wfa *workspaceMongoAccessor) StoreOne(data utils.DBObject) (utils.DBObject, int, error) {
	res, _, err := wfa.Search(nil, data.GetName())
	if err == nil && len(res) > 0 {
		return nil, 409, errors.New("A workspace with the same name already exists")
	}
	d := data.(*Workspace)
	d.DataResources = nil
	d.DatacenterResources = nil
	d.StorageResources = nil
	d.ProcessingResources = nil
	d.WorkflowResources = nil
	return wfa.GenericStoreOne(d, wfa)
}

func (wfa *workspaceMongoAccessor) CopyOne(data utils.DBObject) (utils.DBObject, int, error) {
	return wfa.GenericStoreOne(data, wfa)
}

func (wfa *workspaceMongoAccessor) fill(workflow *Workspace) *Workspace {
	if workflow.Datas != nil && len(workflow.Datas) > 0 {
		dataAccessor := (&data.DataResource{}).GetAccessor(nil)
		for _, id := range workflow.Datas {
			d, _, e := dataAccessor.LoadOne(id)
			if e == nil {
				workflow.DataResources = append(workflow.DataResources, d.(*data.DataResource))
			}
		}
	}
	if workflow.Datacenters != nil && len(workflow.Datacenters) > 0 {
		dataAccessor := (&datacenter.DatacenterResource{}).GetAccessor(nil)
		for _, id := range workflow.Datacenters {
			d, _, e := dataAccessor.LoadOne(id)
			if e == nil {
				workflow.DatacenterResources = append(workflow.DatacenterResources, d.(*datacenter.DatacenterResource))
			}
		}
	}
	if workflow.Storages != nil && len(workflow.Storages) > 0 {
		dataAccessor := (&storage.StorageResource{}).GetAccessor(nil)
		for _, id := range workflow.Storages {
			d, _, e := dataAccessor.LoadOne(id)
			if e == nil {
				workflow.StorageResources = append(workflow.StorageResources, d.(*storage.StorageResource))
			}
		}
	}
	if workflow.Processings != nil && len(workflow.Processings) > 0 {
		dataAccessor := (&processing.ProcessingResource{}).GetAccessor(nil)
		for _, id := range workflow.Processings {
			d, _, e := dataAccessor.LoadOne(id)
			if e == nil {
				workflow.ProcessingResources = append(workflow.ProcessingResources, d.(*processing.ProcessingResource))
			}
		}
	}
	if workflow.Workflows != nil && len(workflow.Workflows) > 0 {
		dataAccessor := (&w.WorkflowResource{}).GetAccessor(nil)
		for _, id := range workflow.Workflows {
			d, _, e := dataAccessor.LoadOne(id)
			if e == nil {
				workflow.WorkflowResources = append(workflow.WorkflowResources, d.(*w.WorkflowResource))
			}
		}
	}
	return workflow
}

func (wfa *workspaceMongoAccessor) LoadOne(id string) (utils.DBObject, int, error) {
	var workflow Workspace
	res_mongo, code, err := mongo.MONGOService.LoadOne(id, wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not retrieve " + id + " from db. Error: " + err.Error())
		return nil, code, err
	}
	res_mongo.Decode(&workflow)

	return wfa.fill(&workflow), 200, nil
}

func (wfa workspaceMongoAccessor) LoadAll() ([]utils.ShallowDBObject, int, error) {
	objs := []utils.ShallowDBObject{}
	res_mongo, code, err := mongo.MONGOService.LoadAll(wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not retrieve any from db. Error: " + err.Error())
		return nil, code, err
	}
	var results []Workspace
	if err = res_mongo.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	for _, r := range results {
		objs = append(objs, wfa.fill(&r))
	}
	return objs, 200, nil
}

func (wfa *workspaceMongoAccessor) Search(filters *dbs.Filters, search string) ([]utils.ShallowDBObject, int, error) {
	objs := []utils.ShallowDBObject{}
	if (filters == nil || len(filters.And) == 0 || len(filters.Or) == 0) && search != "" {
		filters = &dbs.Filters{
			Or: map[string][]dbs.Filter{
				"abstractobject.name": {{Operator: dbs.LIKE.String(), Value: search}},
			},
		}
	}
	res_mongo, code, err := mongo.MONGOService.Search(filters, wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not store to db. Error: " + err.Error())
		return nil, code, err
	}
	var results []Workspace
	if err = res_mongo.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	for _, r := range results {
		objs = append(objs, wfa.fill(&r))
	}
	return objs, 200, nil
}
